/* $Id: fullfactor.gp,v 1.28 2008-01-08 14:13:03 jdemeyer Exp $ */

\r fullfactor_misc
\r fullfactor_bases
\r fullfactor_init
\r fullfactor_cyclo
\r fullfactor_help


/* The main function of this package: The argument F is the object
 * to be factored.  If F is a t_MAT, it is assumed to be a partial
 * factorisation (for example the output of factor(n,0)).  The result
 * will always be factored into coprime factors, so all common factors
 * are put together.  Then each of these factors is factored using
 * factorfunc.  By default this is just PARI/GP's factor() function.  
 * factorfunc may also be a (possible recursive) t_VEC, in that case
 * the factorfuncs are applied from left to right.  See below for some
 * factorfuncs. */
fullfactor(F, factorfunc = factor_default, cmp = cmp_factor_mixed) =
{
	/* We only have one object to be factored */
	if (type(F) != "t_MAT", F = [F;1]~);

	/* The algorithm will modify the given matrix
	 * to contain the full factorisation. */

	/* Loop over all given factorfuncs */
	factorfunc = Vecflat([factorfunc]);
	for (c = 1, #factorfunc,
		/* Loop over all factors in the matrix */
		my(r = 1);  /* Current row */
		while (r <= matsize(F)[1],
			/* Get one factor f^fpow of the old matrix, after this the
			 * r-th row will never be referenced again. */
			my(f = F[r,1]);
			my(fpow = F[r,2]);

			/* Check gcd with later input factors (factor in coprimes) */
			my(i = r+1);
			while(i <= matsize(F)[1],
				my(g = gcd(f, F[i,1]));
				if (g == 1, i++; next); /* No common factor, move on */

				/* We found a non-trivial common factor, good! */
				/* Note that we do NOT increase i in this while loop,
				 * because we may have a higher-degree common factor. */
				my(gpow = fpow + F[i,2]);  /* Exponent of g */

				/* First, divide F[i,1] by g */
				if (F[i,1] == g,
					/* Special case: F[i,1] actually divides f */
					/* Delete row i */
					F = vecextract(F, 1<<(matsize(F)[1])-1 - (1<<(i-1)), [1,2]);
				,
					F[i,1] /= g;
				);

				/* Second, divide f by g and store g for later use */
				if (f == g,
					/* Special case: f is equal to g */
					/* Swap f and g (and forget g) */
					f = g;
					fpow = gpow;
				,
					f /= g;
					F = vconcat(F, [g;gpow]~);  /* Add a row containing g */
				);
			);

			/* Factor the factor coming from row r */
			my(P = simplify(factorfunc[c](f)));
			if (type(P) != "t_MAT", P = [P;1]~);
			
			/* Create a new matrix as follows:
			 *  - First the rows from P (with exponent multiplied by fpow)
			 *  - Then the first r-1 rows from F
			 *  - Then the rows starting from r+1 from F
			 */
			my(Fnew);
			Fnew = matrix(matsize(P)[1] + matsize(F)[1] - 1, 2);

			my(out = 0);
			for (i = 1, matsize(P)[1],
				out++;
				Fnew[out,1] = P[i,1];
				Fnew[out,2] = P[i,2] * fpow;
			);
			for (i = 1, matsize(F)[1],
				if (i == r, next);
				out++;
				Fnew[out,1] = F[i,1];
				Fnew[out,2] = F[i,2];
			);

			F = Fnew;
			r += matsize(P)[1];
		);
	);

	/* Insertion sort the resulting matrix */
	for (i = 1, matsize(F)[1],
		my(f = F[i,1]);
		my(fpow = F[i,2]);
		
		my(k = i-1);
		while (k >= 1 && cmp(f, F[k,1]) < 0,
			F[k+1,1] = F[k,1];
			F[k+1,2] = F[k,2];
			k--;
		);
		if (k+1 < i,
			F[k+1,1] = f;
			F[k+1,2] = fpow;
		);
	);

	F;
}


/* The radical of F (the product of the distinct prime factors of F). */
/* F can be either an object to be factored, or a factorisation.
 *  - If F is not factored, factorfunc will be used to factor it.
 *  - If F is factored, it is assumed to be a full factorisation,
 *    unless factorfunc is explicitly specified.
 */
radical(F, factorfunc = 0) =
{
	/* Should we factor somehow? */
	if (type(factorfunc) == "t_CLOSURE", F = fullfactor(F, factorfunc));
	\\if (type(F) == "t_POL", F = factor_squarefree(F));
	if (type(F) != "t_MAT", F = factor(F));

	prod(i = 1, matsize(F)[1], F[i,1]);
}


/* Various t_CLOSUREs to be used as factorfunc. */

/* Simply call GP's factor function. */
factor_default(x) = factor(x);

/* Trivial factoring function, just factor in coprimes. */
factor_coprime(x) = x;

/* Only look for primes in the prime table. */
factor_trialdiv(x) = factor(x,0);

/* In a factorisation, only keep those factors which are at least primelimit */
factor_above_primelimit(x) = if(type(x) == "t_INT" && x > default(primelimit), x, matrix(0,2));

/* Check whether factors are powers and take roots. */
factor_power(x) =
{
	my(root, pow);
	pow = ispower(x,,&root);
	if (pow, return([root;pow]~));
	[x;1]~;
}

/* Don't factor, just check whether we have primes. */
factor_check(x) =
{
	if (type(x) == "t_INT" && !isprime(x), error("The following factor is NOT prime:\n", x));
	if (type(x) == "t_POL" && !polisirreducible(x), error("The following factor is NOT irreducible:\n", x));
	x;
}

/* Given a vector v of mutually coprime elements, tries to factor x by
 * taking gcd's between x and elements of v.  Returns [F,c] where F is
 * a partial factorisation of x and c is the cofactor such that 
 * gcd(c, v[i]) = 1 for all i = 1, ..., #v.  Note that all the factors
 * in F are divisors of elements of v, but not necessarily equal to
 * elements of v.  For example, if x = 24 and v = [6], then x will be
 * factored as 2^3 * 3.  However, with the same v, x = 36 would be
 * factored as 6^2.  The function factor_gcd() defined below does the
 * same, but in a form that can be used with fullfactor(). */
split_factor_gcd(x,v) =
{
	/* Output */
	my(F = matrix(0,2));

	/* Loop over all elements of v */
	for (i = 1, #v,
		my(g = gcd(x, v[i]));
		if (g == 1, next);

		my(n = 0);
		while (g != 1,
			/* Find maximal n such that g^n divides x */
			while ((x % g) == 0,
				n++;
				x /= g;
			);

			my(gnew = gcd(x,g));

			/* Store cofactor g/gnew in factorisation */
			F = vconcat(F, [g/gnew;n]~);

			/* Continue with the factor gnew, note that we keep the exponent n */
			g = gnew;
		);
	);

	[F,x];
}

/* Same as split_factor_gcd() above, but returns a factorisation
 * matrix (with the possible cofactor appended).  Note that this is a
 * function with one argument, returning a t_CLOSURE.  So it can be
 * used as fullfactor(24, factor_gcd([2,7])); */
factor_gcd(v) = 
{
	if (type(v) != "t_VEC", v = [v]);
	(x->
		my(P = split_factor_gcd(x,v));
		if (P[2] != 1,
			/* Cofactor != 1, append it to P[1] */
			P[1] = vconcat(P[1], [P[2];1]~);
		);
		P[1];
	);
}


/* Factor a polynomial into squarefree factors */
factor_squarefree(f) =
{
	if (type(f) != "t_POL", return(x));

	my(r = f/gcd(f,f'));

	return(factor_squarefree_recurse(f, r, 1, 1, poldegree(f)));
}

factor_squarefree_recurse(f,r,d,m,maxd) =
{
	/* f = polynomial to be factored */
	/* r = product of all irreducible factors in f whose valuation in f
	       is not a multiple of the characteristic. */
	/* d = degree of the factor that we will find */
	/* m = multiplier of degree (this is a power of the characteristic) */
	/* maxd = maximum d occuring in f */

	if (!poldegree(f), return(matrix(0,2)));

	maxd = min(maxd, poldegree(f) - poldegree(r));
	if (default(debug) >= 3,
		print("factor_squarefree: deg(f) = ", poldegree(f), "; deg(r) = ", poldegree(r), "; maxd = ", maxd)
	);

	while (!poldegree(r),
		/* We must be having a p-th power in characteristic p */
		/* First, figure out the characteristic */
		my(p = pollead(f));
		if (type(p) == "t_INTMOD", p = p.mod);
		if (type(p) == "t_POLMOD", p = gcd(apply(z->z.mod,Vec(p.mod))));
		if (type(p) == "t_FFELT", p = p.p);

		if (type(p) != "t_INT", error("factor_squarefree is not implemented for polynomials of type ", type(p)));
		if (!ispseudoprime(p), error("factor_squarefree: characteristic ", p, " is not prime!"));
	
		/* Replace x^p by x in f */
		my(v = vector(poldegree(f)/p + 1, i, polcoeff(f, (i-1)*p)));

		/* For every element in f, take the p-th root */
		for (i = 1, #v,
			my(a = v[i]);
			if (!a, next);
			if (type(a) == "t_INTMOD", next);
			if (type(a) == "t_POLMOD", v[i] = a^(p^(poldegree(a.mod)-1)); next);
			v[i] = sqrtn(a,p);
		);

		/* p-th root of f */
		f = Polrev(v, variable(f));

		r = f/gcd(f,f');
		d = 1; m *= p; maxd = poldegree(f) - poldegree(r);

		if (default(debug) >= 3,
			print("\t>>> deg(f) = ", poldegree(f), "; deg(r) = ", poldegree(r), "; maxd = ", maxd)
		);
	);

	if (maxd <= 4,
		/* Direct computation, return factorisation F */
		my(F = matrix(0,2));

		while (poldegree(r),
			f /= r;

			/* Find factor of degree d in f */
			if (f % r,  /* Otherwise, r doesn't change and there is no factor of degree d */
				my(g = gcd(r, f));
				F = vconcat(F, [r/g; d*m]~);
				r = g;
			);
			d++;
		);

		if (poldegree(f), F = vconcat(F, factor_squarefree_recurse(f,r,d,m,maxd)));
		return(F);
	);

	gettime();

	my(split = sqrtint(maxd));
	my(fsplit = gcdmod(Mod(r,f)^split));  /* gcd(f, r^split) */

	/* Compute low and high parts */
	my(flow = 1, rlow = 1, fhigh = 1, rhigh = 1);

	if (poldegree(f) == poldegree(fsplit),
		/* Only low part */
		rlow = r;
		flow = f;
	,
		/* At least some high part */
		fhigh = f/fsplit;
		rhigh = gcd(r, fhigh);
		if (poldegree(r) != poldegree(rhigh),
			/* Also low part */
			rlow = r/rhigh;
			flow = fsplit/(rhigh ^ split);
		);
	);

	if (default(debug) >= 3,
		print("\tsplit = ", split, "; deg(flow) = ", poldegree(flow), "; deg(fhigh) = ", poldegree(fhigh), "; time: ", gettime(), " ms.")
	);

	if (!poldegree(fhigh), return(factor_squarefree_recurse(flow, rlow, d, m, split)));
	if (!poldegree(flow), return(factor_squarefree_recurse(fhigh, rhigh, d + split, m, maxd - split)));

	return(vconcat(
		factor_squarefree_recurse(flow, rlow, d, m, split);
	,
		factor_squarefree_recurse(fhigh, rhigh, d + split, m, maxd - split);
	));
}


/* Find cyclotomic factors, i.e. factors of the form polcyclo(e,b) with
 * e and b "small".  factor_cyclo does trial division before calling
 * factor_cyclo1(), defined in fullfactor_cyclo.gp. */
factor_cyclo(x) = fullfactor(x, [factor_trialdiv, factor_cyclo1]);


/* Compare two elements in a factorisation. */
cmp_factor_mixed(x,y) =
{
	/* If types differ, compare types */
	my(tx = x, ty = y);
	while (1,
		if (type(tx) == type(ty),
			/* Both are t_POL: recurse */
			if (type(tx) == "t_POL",
				tx = pollead(tx); ty = pollead(ty);
				next;
			);

			/* Equal types, but not t_POL: go on */
			break;
		);

		/* One t_POL: the t_POL comes last */
		if (type(ty) == "t_POL", return(-1));
		if (type(tx) == "t_POL", return(1));

		/* No t_POL but different type: compare types lexicographically */
		if (type(tx) < type(ty), return(-1), return(1));
	);

	/* x and y are both polynomials: compare degree, then sort
	 * lexicographically (by recursing). */
	if (type(x) == "t_POL",
		my(d = gcmp(poldegree(x), poldegree(y)));
		if (d != 0, return(d));

		forstep (i = poldegree(x), 0, -1,
			d = cmp_factor_mixed(polcoeff(x,i), polcoeff(y,i));
			if (d != 0, return(d));
		);

		/* All coefficients compare equal */
		return(0);
	);

	/* Other types */
	if (type(x) == "t_INTMOD" || type(x) == "t_POLMOD",
		return (cmp_factor_mixed(lift(x), lift(y)));
	);

	if (type(x) == "t_INT" || type(x) == "t_FRAC" || type(x) == "t_REAL",
		return (gcmp(x,y));
	);

	if (type(x) == "t_PADIC",
		return (cmp_padic(x,y));
	);

	/* No sensible comparison possible */
	0;
};

cmp_factor_none(x,y) = 0;
