/* $Id: fullfactor_cyclo.gp,v 1.18 2009-11-24 11:36:47 jdemeyer Exp $ */


/* Accessor functions for G[k] in factor_cyclo1() */
f.factor = f[1];
f.power = f[2];
f.base = f[3];
f.expdiv = f[4];

/* Find cyclotomic factors of x.  This function should only be used
 * AFTER TRIAL DIVISION up to 500000, see factor_cyclo() in
 * fullfactor.gp.
 *
 * If flag == 1, return a 4-column factorisation matrix instead, where
 * every row is [factor, power, base, exponent].  Here, factor and
 * power are the usual two columns, and factor is a divisor of
 * polcyclo(exponent, base).  For non-cyclotomic factors, base and
 * exponent will be 0. */
factor_cyclo1(x, flag = 0) =
{
	/* Check that primelimit >= 500000 */
	if (default(primelimit) < 500000,
		error("factor_cyclo1: primelimit must be at least 500000.  Please type default(primelimit, 500000) and try again.");
	);

	my(DEBUG = default(debug));
	my(fourcolumn = bitand(flag, 1));

	/* Some obvious cases where there's nothing to do */
	if (!fourcolumn,
		if (type(x) != "t_INT", return([x;1]~));
		if (x == 1, return(matrix(0,2)));
		if (x < 500000^2, return([x;1]~));  /* x is prime if we trial divided up to 500000 */
	,
		if (type(x) != "t_INT", return([x;1;0;0]~));
		if (x == 1, return(matrix(0,4)));
	);

	/* Save x to sanity check at the end that the factorisation is correct */
	my(x0 = x);

	/* Number of prime/composite/total bases */
	my(numpbases = #factor_cyclo_baseprime);
	my(numcbases = #factor_cyclo_basecomp);
	my(numbases = numpbases + numcbases);

	/* B1 and B2 bounds for p-1 */
	my(B1 = log(x) / (4 + numpbases));
	my(B2 = 100.0 * log(x) / (4 + numbases));
	if (B2 <= 2399, B2 = 2399, B2 = ceil(B2));
	if (B1 * B1 < B2, B1 = ceil(sqrt(B2)), B1 = ceil(B1));
	B1 = precprimepower(B1);
	B2 = precprimepower(B2);

	/* b^i mod x for all prime bases b + extra registers for composite bases */
	my(bexp = vector(numpbases + factor_cyclo_baseregs));
	for (i = 1, numpbases, bexp[i] = Mod(factor_cyclo_baseprime[i],x));

	/* Saved data for rewinding */
	my(bexprew, nrew, prew);

	/* Default number of loops to be accumulated,
	 * actual number of loops to be accumulated,
	 * loops already accumulated. */
	my(defaultaccum, numaccum, accum);


	/* The factoring will proceed in 3 stages: stages 1 and 2 are just
	 * like the usual p-1 factoring method. (except that we try specific
	 * bases).  In stage 3 the factors found will be identified as
	 * divisors of polcyclo(n,b) and further factored (Aurifeuillian and
	 * data lookup). */

	/* Save factors found in stages 1 and 2.  Whenever we find a factor,
	 * we remove it from x.
	 * G is a vector with one component for every factor:
	 *  - G[k].factor: the factor itself.
	 *  - G[k].power: the power with which the factor appears in x.
	 *  - G[k].base: the base such that G[k].factor divides G[k].base^n - 1
	 *    for some n.
	 *  - G[k].expdiv: a sure divisor of the exponent.
	 */
	my(G = vector(0));


	/* Stage 1 */
	if (DEBUG >= 2, print("factor_cyclo1: STAGE 1: B1 = ", B1); gettime());

	/* How many products to accumulate? */
	numaccum = defaultaccum = ceil(48/numpbases - 32/numbases);
	accum = 0;

	/* Save variables for rewinding */
	bexprew = bexp; nrew = 1;
	/* Loop up to B2, but do only powers of 2 from B1 */
	for (n = 2, B2,
		my(p = isprimepower(n));
		if (n > B1 && (p != 2), next);
		if (!p, next);

		if (DEBUG >= 4, print("n = ", n, ": p = ", p));

		for (i = 1, numpbases,
			bexp[i] = bexp[i]^p;
		);

		/* B1 bound reached, finish */
		if (n >= B1, numaccum = accum+1);
		if (accum++ >= numaccum,
			/* Compute a := accumulated product */
			my(a = Mod(1,x));
			my(reg = 1, c = 0);  /* register, composite base index (0 while prime) */
			while (1,
				if (reg > numpbases,
					/* Compute b^n for composite base */
					if (c++ > numcbases, break);
					bexp[reg] = bexp[factor_cyclo_basecalc[2*c-1]] * bexp[factor_cyclo_basecalc[2*c]];
				);
				a *= bexp[reg] - 1;
				if (reg++ > #bexp, reg -= factor_cyclo_baseregs);
			);

			my(g = gcdmod(a));
			if (g > 1,
				if (accum > 1,  /* Try again with a lower numaccum */
					n = nrew; bexp = bexprew;
					numaccum >>= 2;
					accum = 0;
					if (DEBUG >= 4, print("<<<"));
					next;
				);

				/* Try all bases again and check gcd for every base */
				my(reg = 1, c = 0);  /* register, composite base index (0 while prime) */
				while (1,
					if (reg > numpbases,
						/* Compute b^n for composite base */
						if (c++ > numcbases, break);
						bexp[reg] = bexp[factor_cyclo_basecalc[2*c-1]] * bexp[factor_cyclo_basecalc[2*c]];
					);
					my(h = gcd(lift(bexp[reg] - 1), g));
					if (h > 1,
						/* We found a factor */
						my(base = if(reg <= numpbases, factor_cyclo_baseprime[reg], factor_cyclo_basecomp[c]));
						
						/* Check how this factor h fits in x */
						my(P = split_factor_gcd(x , [h]));
						x = P[2];  /* Cofactor */
						P = P[1];
						for (j = 1, matsize(P)[1],
							/* Factor to add to G */
							G = concat(G, [[P[j,1], P[j,2], base, p]]);
							if (DEBUG >= 3, print("    Temp factor ", P[j,1], if(P[j,2]>1, Str("^", P[j,2]), "") , " in ", base, "^", p, "d - 1"));
						);

						g /= h;
						if (g == 1, break);
					);
					if (reg++ > #bexp, reg -= factor_cyclo_baseregs);
				);

				/* Adjust moduli in bexp for found factor */
				if (x == 1, break);
				for (i = 1, numpbases, bexp[i] %= x);

				numaccum = defaultaccum;
			);

			/* Save values for rewinding and restart accumulating */
			bexprew = bexp; nrew = n;
			accum = 0;
		);
	);
	if (DEBUG >= 2, print("factor_cyclo1: Finished stage 1 in ", gettime(), " ms"));


	/* Stage 2, only if x > 1 */
	if (x > 1,
	if (DEBUG >= 2, print("factor_cyclo1: STAGE 2: B2 = ", B2); gettime());

	/* For every prime base b, save b^2L, b^4L, ..., b^24L % x,
	 * where b^L is the result of stage 1. */
	my(MAXGAP = 12);
	my(bexpmuls = matrix(numpbases, MAXGAP));
	for (i = 1, numpbases,
		bexpmuls[i,1] = lift(bexp[i]^2);
		for (j = 2, MAXGAP,
			bexpmuls[i,j] = (bexpmuls[i,j-1] * bexpmuls[i,1]) % x;
		);
	);

	/* How many products to accumulate? */
	numaccum = defaultaccum = ceil(16/numbases);
	accum = 0;
	my(a = 1);  /* accumulated product */

	/* Save variables for rewinding */
	bexprew = bexp; prew = 1;
	/* Loop up to B2 */
	my(prevp = prew);
	forstep (p = 3, B2, 2,
		if (!isprimepower(p), next);

		my(gap = (p - prevp) >> 1);
		if (DEBUG >= 4, print("n = ", p, "; gap = ", gap << 1));
		while (gap,
			my(j = min(gap, MAXGAP)); gap -= j;
			for (i = 1, numpbases, bexp[i] *= bexpmuls[i,j]);
		);
		prevp = p;

		/* Accumulate */
		my(reg = 1, c = 0);  /* register, composite base index (0 while prime) */
		while (1,
			if (reg > numpbases,
				/* Compute b^n for composite base */
				if (c++ > numcbases, break);
				bexp[reg] = bexp[factor_cyclo_basecalc[2*c-1]] * bexp[factor_cyclo_basecalc[2*c]];
			);
			a *= bexp[reg] - 1;
			if (reg++ > #bexp, reg -= factor_cyclo_baseregs);
		);

		/* B2 bound reached, finish */
		if (p >= B2, numaccum = accum+1);

		if (accum++ >= numaccum,
			my(g = gcdmod(a));
			if (g > 1,
				if (accum > 1,  /* Try again with numaccum == 1 */
					p = if (prew == 1, 3, prew); bexp = bexprew;
					prevp = prew;
					numaccum = 1;
					accum = 0; a = 1;
					if (DEBUG >= 4, print("<<<"));
					next;
				);

				/* Try all bases again and check gcd for every base */
				my(reg = 1, c = 0);  /* register, composite base index (0 while prime) */
				while (1,
					if (reg > numpbases,
						/* Compute b^n for composite base */
						if (c++ > numcbases, break);
						bexp[reg] = bexp[factor_cyclo_basecalc[2*c-1]] * bexp[factor_cyclo_basecalc[2*c]];
					);
					my(h = gcd(lift(bexp[reg] - 1), g));
					if (h > 1,
						/* We found a factor */
						my(base = if(reg <= numpbases, factor_cyclo_baseprime[reg], factor_cyclo_basecomp[c]));
						
						/* Check how this factor h fits in x */
						my(P = split_factor_gcd(x , [h]));
						x = P[2];  /* Cofactor */
						P = P[1];
						for (j = 1, matsize(P)[1],
							/* Factor to add to G */
							G = concat(G, [[P[j,1], P[j,2], base, p]]);
							if (DEBUG >= 3, print("    Temp factor ", P[j,1], if(P[j,2]>1, Str("^", P[j,2]), "") , " in ", base, "^", p, "d - 1"));
						);

						g /= h;
						if (g == 1, break);
					);
					if (reg++ > #bexp, reg -= factor_cyclo_baseregs);
				);

				/* Adjust moduli in bexp and bexpmuls for found factor */
				if (x == 1, break);
				for (i = 1, numpbases,
					bexp[i] %= x;
					for (j = 1, MAXGAP, bexpmuls[i,j] %= x);
				);

				numaccum = defaultaccum;
			);

			/* Save values for rewinding and restart accumulating */
			bexprew = bexp; prew = p;
			accum = 0; a = 1;
		);
	);
	if (DEBUG >= 2, print("factor_cyclo1: Finished stage 2 in ", gettime(), " ms"));
	);  /* if (x > 1) */
	

	/* Save factorisation in F.
	 * Start with unfactored part of x (after this, we forget x) */
	my(F);
	if (x > 1,
		F = if(fourcolumn, [x;1;0;0], [x;1])~;
		x0 /= x;
		if (DEBUG >= 2, print(" ** Cofactor ", x));
	,
		F = matrix(0, if(fourcolumn, 4, 2));
	);

	/* Stage 3, identify the factors that we found. */
	if (#G,
	if (DEBUG >= 2, print("factor_cyclo1: STAGE 3: ", #G, " factor", if(#G>1,"s",""), " to identify"); gettime());

	/* Loop over all factors.  These factors will be removed from G,
	 * so we loop until G is empty. */
	while (#G,
		/* Get data from G[1] and delete that entry */
		x = G[1].factor;
		my(xpow = G[1].power);
		my(base = G[1].base);
		my(e = G[1].expdiv);
		bexp = Mod(base, x) ^ e;  /* Just one element for the correct base */
			
		G = vecextract(G, (1<<#G) - 2);

		/* Extract factors */
		if (gcdmod(bexp - 1) > 1,
			my(P = split_factor_cyclo(x, base, e));
			x = P[2];  /* Cofactor */
			P = P[1];

			/* Add these factors to F */
			for (j = 1, matsize(P)[1],
				P[j,2] *= xpow;
				if (fourcolumn,
					F = vconcat(F, [P[j,1]; P[j,2]; base; e]~);
				,
					F = vconcat(F, [P[j,1]; P[j,2]]~);
				);
				if (gcd(x0, P[j,1]^(P[j,2] + 1)) != P[j,1]^P[j,2],
					error("factor_cyclo1: factor ", P[j,1], "^", P[j,2], " of given x is not correct.  This is a bug, please report!");
				);
				x0 /= P[j,1]^P[j,2];
				if (DEBUG >= 3, print(" ** Final factor ", P[j,1], if(P[j,2]>1, Str("^", P[j,2]), "") , " in ", base, "^", e, " - 1"));
			);

			if (x == 1, next);
		);

		if (DEBUG >= 3, print("    Processing factor ", x, " in ", base, "^", e, "d - 1"));

		/* How many products to accumulate? */
		numaccum = defaultaccum = 16;
		accum = 0;

		/* Save variables for rewinding */
		bexprew = bexp; nrew = 1;
		/* Loop forever starting with n = 2 */
		my(n = 1); while (n++; 1,
			my(p = isprimepower(n));
			if (!p, next);

			if (DEBUG >= 4, print("n = ", n, ": p = ", p));

			bexp = bexp^p;

			if (accum++ >= numaccum,
				my(g = gcdmod(bexp - 1));
				if (g > 1,
					if (accum > 1,  /* Try again with a lower numaccum */
						n = nrew; bexp = bexprew;
						numaccum >>= 2;
						accum = 0;
						if (DEBUG >= 4, print("<<<"));
						next;
					);

					/* We found a factor */
					/* Check how this factor g fits in x */
					my(P = split_factor_gcd(x , [g]));
					x = P[2];  /* Cofactor */
					P = P[1];
					for (j = 1, matsize(P)[1],
						/* Factor to add to G */
						G = concat(G, [[P[j,1], P[j,2]*xpow, base, p*e]]);
						if (DEBUG >= 3, print("        Sub factor ", P[j,1], if(P[j,2]>1, Str("^", P[j,2]), "") , " in ", base, "^", p*e, "d - 1"));
					);

					/* Adjust moduli in bexp for found factor */
					if (x == 1, break);
					bexp %= x;

					numaccum = defaultaccum;
				);

				/* Save values for rewinding and restart accumulating */
				bexprew = bexp; nrew = n;
				accum = 0;
			);
		);
	);
	if (DEBUG >= 2, print("factor_cyclo1: Finished stage 3 in ", gettime(), " ms"));
	);  /* if (#G) */

	if (x0 != 1, error("factor_cyclo1: Factor ", x0, " not accounted for.  This is a bug, please report!"));
	F;
}


/* Given a number x, extract all factors of polcyclo(e,b) by doing
 * database lookup and Aurifeuillian factorisation.  Returns [F,c]
 * just like split_factor_gcd().  Here F is a partial factorisation
 * of x of all factors in gcd(x, b^e-1) x and c is the cofactor. */
split_factor_cyclo(x,b,e) =
{
	/* Relevant cyclotomic part: gcd(b^e-1, x) */
	my(g = gcdmod(Mod(b,x)^e - 1));

	/* Collect possible factors of polcyclo(e,b) in vec_factors */
	my(vec_factors = vector(0));


	/* 1) Read factors from file */
	if (type(fullfactor_datadir) == "t_STR" && fullfactor_datadir != "" && b <= 12,
		filename = Str(fullfactor_datadir, "cyclo", b, ".", e%9);
		my(vec_file = readvec(filename));

		/* Get all the factors of polcyclo(b,e)
		 * from vec_file, store in vec_factors. */
		for (i = 1, #vec_file,
			if (vec_file[i][1] == b && vec_file[i][2] == e,
				vec_factors = concat(vec_factors, vec_file[i][3]);
			);
		);
	);


	/* 2) Aurifeuillian factorisation */
	/* Write b = bsqrfree * bsqr^2 */
	my(bsqr = 1, bsqrfree = 1);
	my(B = factor(b));
	for (i = 1, matsize(B)[1], bsqr *= B[i,1]^(B[i,2]\2); if (B[i,2]%2, bsqrfree *= B[i,1]));

	/* aurifmod = bsqrfree, but multiplied by 2 if it is 2 or 3 mod 4. */
	my(aurifmod = bsqrfree);
	if (aurifmod % 4 >= 2, aurifmod *= 2);

	/* Do we have an Aurifeuillian factorisation? */
	if ((e + aurifmod) % (2*aurifmod) == 0,
		my(aurif_arg, aurif_fac);
		aurif_arg = bsqr * Mod(b,g)^((e - aurifmod)/(2*aurifmod));
		aurif_fac = gcdmod(polaurif(bsqrfree, aurif_arg));
		
		/* Only add non-trivial factorisation to vec_factors */
		if (aurif_fac > 1 && aurif_fac < g,
			vec_factors = concat(vec_factors, aurif_fac);
			if (default(debug) >= 3,
				print("    Aurifeuillian factorisation ", aurif_fac, " * ", g/aurif_fac);
			);
		);
	);


	/* 3) Add g */
	vec_factors = concat(vec_factors, g);


	/* Now factor x as much as possible using the factor(s) found */
	split_factor_gcd(x, vec_factors);
}


/* Returns partial factorisation of b^n - 1. */
partial_factor_pn_1(b,n) =
{
	my(F = matrix(0,2));
	my(P);
	my(trialdivbound = default(primelimit)^2);

	fordiv(n, d,
		my(x = polcyclo(d, b));
		if (x^2 <= trialdivbound,
			/* x will be completely factored by trial division */
			P = [x;1]~;
		,
			P = split_factor_cyclo(x, b, d)[1];
		);
		F = vconcat(F, P);
	);
	fullfactor(F, factor_trialdiv);
}

/* Returns partial factorisation of b^n + 1. */
partial_factor_pn_p1(b,n) =
{
	my(F = matrix(0,2));
	my(P);
	my(trialdivbound = default(primelimit)^2);

	/* Write 2*n as n * n_even, where n is odd and n_even a power of 2 */
	my(n_even = 2);
	while ((n % 2) == 0, n >>= 1; n_even <<= 1);

	fordiv(n, d,
		my(x = polcyclo(d*n_even, b));
		if (x^2 <= trialdivbound,
			/* x will be completely factored by trial division */
			P = [x;1]~;
		,
			P = split_factor_cyclo(x, b, d*n_even)[1];
		);
		F = vconcat(F, P);
	);
	fullfactor(F, factor_trialdiv);
}


polaurif(b,x='x) =
{
	/* This computation uses t_REALs which are then rounded to t_INTs.
	 * This needs a sufficiently high precision.  This precision below
	 * has been verified to work for all b <= 4547. */
	my(savedprec = default(realprecision));
	default(realprecision, round(sqrt(2*b)) + 28);

	/* Vector to compute products. Every element a[i+1] contains the
	 * product of 2^i quadratic factors of the polynomial.  The total
	 * number of factors to be computed is equal to eulerphi(2b)/2,
	 * the degree of the polynomial is eulerphi(2b). */
	a = vector(logint(eulerphi(2*b),2));

	/* All elements in (Z/2bZ)^* / {1,-1} */
	my(g = vectorsmall(eulerphi(2*b)/2));
	my(c = sqrtint(b)); while (gcd(b,c) > 1, c++);
	my(n = 0);  /* Counter, how many elements in g so far? */
	forstep (i = 1, b, 2,
		if (gcd(i,b) != 1, next);
		n++; g[n] = (i*c) % b;
	);

	/* Evaluate polaurif(b, t = x/sqrt(b)) which is monic */
	my(t = 't); /* Polynomial variable */
	if (b%4 == 1,
		/* b = 1 mod 4 */
		for (n = 1, #g,
			my(k = g[n]);
			my(val2 = valuation(n, 2));            /* Where to put this factor? */
			a[val2+1] = t^2 + 2*kronecker(k,b)*cos(2*k/b*Pi)*t + 1;
			for (i = 1, val2, a[val2+1] *= a[i]);  /* Multiply with earlier factors */
		);
	);
	
	if ((b%4) == 3,
		/* b = 3 mod 4 */
		for (n = 1, #g,
			my(k = g[n]);
			my(val2 = valuation(n, 2));            /* Where to put this factor? */
			a[val2+1] = t^2 + 2*kronecker(k,b)*sin(2*k/b*Pi)*t + 1;
			for (i = 1, val2, a[val2+1] *= a[i]);  /* Multiply with earlier factors */
		);
	);

	if ((b%8) == 2,
		/* b = 2 mod 8 */
		for (n = 1, #g,
			my(k = g[n]);
			my(val2 = valuation(n, 2));            /* Where to put this factor? */
			a[val2+1] = t^2 - 2*sin(4*k/b*Pi + kronecker(k,b/2)*Pi/4)*t + 1;
			for (i = 1, val2, a[val2+1] *= a[i]);  /* Multiply with earlier factors */
		);
	);

	if ((b%8) == 6,
		/* b = 6 mod 8 */
		for (n = 1, #g,
			my(k = g[n]);
			my(val2 = valuation(n, 2));            /* Where to put this factor? */
			if (k < b/2,
				a[val2+1] =	t^2 - 2*cos(4*k/b*Pi - kronecker(k,b/2)*Pi/4)*t + 1;
			,
				a[val2+1] = t^2 + 2*cos(4*k/b*Pi + kronecker(k,b/2)*Pi/4)*t + 1;
			);
			for (i = 1, val2, a[val2+1] *= a[i]);  /* Multiply with earlier factors */
		);
	);

	/* Final multiplication */
	for (i = 1, #a-1,
		/* Only if bit i-1 is set in #g */
		if ((#g >> (i-1)) % 2, a[#a] *= a[i]);
	);
	a = a[#a];

	/* Compute r = a(sqrt(b)*x) */
	r = 0;
	my(sqrtb = sqrt(b));

	/* To substitute, a_i must be multiplied with s^i */
	/* We do this by multiplying with b^(i\2) * sqrt(b)^(i%2) */
	/* In the mean time, also compute the error (maximum norm) */
	my(maxerr = 0);
	forstep (i = poldegree(a), 0, -1,
		my(c = polcoeff(a,i));  /* Coefficient of t^i in a(t) */
		if ((i%2) == 1, c *= sqrtb; r *= b);
		
		/* Round */
		my(k = round(c));
		if (abs(c-k) > maxerr, maxerr = abs(c-k));
		
		r = r*x + k;
	);

	/* Sanity check the error */
	if (default(debug) >= 4, print("polaurif(", b, "): error = ", maxerr));
	if (maxerr > 1e-3, error("polaurif(", b, "): Insufficient precision"));

	default(realprecision, savedprec);

	r;
}
